import threading  # noqa

from typing import Callable, Optional  # noqa
import watchdog.observers  # pylint: disable=import-error
from watchdog import events  # pylint: disable=import-error

from chalice.cli.filewatch import FileWatcher, WorkerProcess


class WatchdogWorkerProcess(WorkerProcess):
    """Worker that runs the chalice dev server."""

    def _start_file_watcher(self, project_dir):
        # type: (str) -> None
        restart_callback = WatchdogRestarter(self._restart_event)
        watcher = WatchdogFileWatcher()
        watcher.watch_for_file_changes(
            project_dir, restart_callback)


class WatchdogFileWatcher(FileWatcher):
    def watch_for_file_changes(self, root_dir, callback):
        # type: (str, Callable[[], None]) -> None
        observer = watchdog.observers.Observer()
        observer.schedule(callback, root_dir, recursive=True)
        observer.start()


class WatchdogRestarter(events.FileSystemEventHandler):

    def __init__(self, restart_event):
        # type: (threading.Event) -> None
        # The reason we're using threading
        self.restart_event = restart_event

    def on_any_event(self, event):
        # type: (events.FileSystemEvent) -> None
        # If we modify a file we'll get a FileModifiedEvent
        # as well as a DirectoryModifiedEvent.
        # We only care about reloading is a file is modified.
        if event.is_directory:
            return
        self()

    def __call__(self):
        # type: () -> None
        self.restart_event.set()




# This is the version that's written to the config file
# on a `chalice new-project`.  It's also how chalice is able
# to know when to warn you when changing behavior is introduced.
CONFIG_VERSION = '2.0'


TEMPLATE_APP = """\
from chalice import Chalice

app = Chalice(app_name='%s')


@app.route('/')
def index():
    return {'hello': 'world'}


# The view function above will return {"hello": "world"}
# whenever you make an HTTP GET request to '/'.
#
# Here are a few more examples:
#
# @app.route('/hello/{name}')
# def hello_name(name):
#    # '/hello/james' -> {"hello": "james"}
#    return {'hello': name}
#
# @app.route('/users', methods=['POST'])
# def create_user():
#     # This is the JSON body the user sent in their POST request.
#     user_as_json = app.current_request.json_body
#     # We'll echo the json body back to the user in a 'user' key.
#     return {'user': user_as_json}
#
# See the README documentation for more examples.
#
"""


GITIGNORE = """\
.chalice/deployments/
.chalice/venv/
"""

DEFAULT_STAGE_NAME = 'dev'
DEFAULT_APIGATEWAY_STAGE_NAME = 'api'
DEFAULT_ENDPOINT_TYPE = 'EDGE'
DEFAULT_TLS_VERSION = 'TLS_1_2'

DEFAULT_LAMBDA_TIMEOUT = 60
DEFAULT_LAMBDA_MEMORY_SIZE = 128
MAX_LAMBDA_DEPLOYMENT_SIZE = 50 * (1024 ** 2)
# This is the name of the main handler used to
# handle API gateway requests.  This is used as a key
# in the config module.
DEFAULT_HANDLER_NAME = 'api_handler'

MIN_COMPRESSION_SIZE = 0
MAX_COMPRESSION_SIZE = 10485760

LAMBDA_TRUST_POLICY = {
    "Version": "2012-10-17",
    "Statement": [{
        "Sid": "",
        "Effect": "Allow",
        "Principal": {
            "Service": "lambda.amazonaws.com"
        },
        "Action": "sts:AssumeRole"
    }]
}


CLOUDWATCH_LOGS = {
    "Effect": "Allow",
    "Action": [
        "logs:CreateLogGroup",
        "logs:CreateLogStream",
        "logs:PutLogEvents"
    ],
    "Resource": "arn:*:logs:*:*:*"
}


VPC_ATTACH_POLICY = {
    "Effect": "Allow",
    "Action": [
        "ec2:CreateNetworkInterface",
        "ec2:DescribeNetworkInterfaces",
        "ec2:DetachNetworkInterface",
        "ec2:DeleteNetworkInterface"
    ],
    "Resource": "*"
}

XRAY_POLICY = {
    'Effect': 'Allow',
    'Action': [
        'xray:PutTraceSegments',
        'xray:PutTelemetryRecords',
    ],
    'Resource': '*'
}

CODEBUILD_POLICY = {
    "Version": "2012-10-17",
    # This is the policy straight from the console.
    "Statement": [
        {
            "Action": [
                "logs:CreateLogGroup",
                "logs:CreateLogStream",
                "logs:PutLogEvents"
            ],
            "Resource": "*",
            "Effect": "Allow"
        },
        {
            "Action": [
                "s3:GetObject",
                "s3:GetObjectVersion",
                "s3:PutObject"
            ],
            "Resource": "arn:*:s3:::*",
            "Effect": "Allow"
        }
    ]
}

CODEPIPELINE_POLICY = {
    "Version": "2012-10-17",
    # Also straight from the console setup.
    "Statement": [
        {
            "Action": [
                "s3:GetObject",
                "s3:GetObjectVersion",
                "s3:GetBucketVersioning",
                "s3:CreateBucket",
                "s3:PutObject",
                "s3:PutBucketVersioning"
            ],
            "Resource": "*",
            "Effect": "Allow"
        },
        {
            "Action": [
                "codecommit:CancelUploadArchive",
                "codecommit:GetBranch",
                "codecommit:GetCommit",
                "codecommit:GetUploadArchiveStatus",
                "codecommit:UploadArchive"
            ],
            "Resource": "*",
            "Effect": "Allow"
        },
        {
            "Action": [
                "cloudwatch:*",
                "iam:PassRole"
            ],
            "Resource": "*",
            "Effect": "Allow"
        },
        {
            "Action": [
                "lambda:InvokeFunction",
                "lambda:ListFunctions"
            ],
            "Resource": "*",
            "Effect": "Allow"
        },
        {
            "Action": [
                "cloudformation:CreateStack",
                "cloudformation:DeleteStack",
                "cloudformation:DescribeStacks",
                "cloudformation:UpdateStack",
                "cloudformation:CreateChangeSet",
                "cloudformation:DeleteChangeSet",
                "cloudformation:DescribeChangeSet",
                "cloudformation:ExecuteChangeSet",
                "cloudformation:SetStackPolicy",
                "cloudformation:ValidateTemplate",
                "iam:PassRole"
            ],
            "Resource": "*",
            "Effect": "Allow"
        },
        {
            "Action": [
                "codebuild:BatchGetBuilds",
                "codebuild:StartBuild"
            ],
            "Resource": "*",
            "Effect": "Allow"
        }
    ]
}


WELCOME_PROMPT = r"""

   ___  _  _    _    _     ___  ___  ___
  / __|| || |  /_\  | |   |_ _|/ __|| __|
 | (__ | __ | / _ \ | |__  | || (__ | _|
  \___||_||_|/_/ \_\|____||___|\___||___|


The python serverless microframework for AWS allows
you to quickly create and deploy applications using
Amazon API Gateway and AWS Lambda.

Please enter the project name"""


MISSING_DEPENDENCIES_TEMPLATE = r"""
Could not install dependencies:
%s
You will have to build these yourself and vendor them in
the chalice vendor folder.

Your deployment will continue but may not work correctly
if missing dependencies are not present. For more information:
http://aws.github.io/chalice/topics/packaging.html

"""


EXPERIMENTAL_ERROR_MSG = """

You are using experimental features without explicitly opting in.
Experimental features do not guarantee backwards compatibility and may be
removed in the future.  If you'd still like to use these experimental features,
you can opt in by adding this to your app.py file:\n\n%s

See https://aws.github.io/chalice/topics/experimental.html for more
details.
"""


SQS_EVENT_SOURCE_POLICY = {
    "Effect": "Allow",
    "Action": [
        "sqs:ReceiveMessage",
        "sqs:DeleteMessage",
        "sqs:GetQueueAttributes",
    ],
    "Resource": "*",
}


KINESIS_EVENT_SOURCE_POLICY = {
    "Effect": "Allow",
    "Action": [
        "kinesis:GetRecords",
        "kinesis:GetShardIterator",
        "kinesis:DescribeStream",
        "kinesis:ListStreams",
    ],
    "Resource": "*",
}


DDB_EVENT_SOURCE_POLICY = {
    "Effect": "Allow",
    "Action": [
        "dynamodb:DescribeStream",
        "dynamodb:GetRecords",
        "dynamodb:GetShardIterator",
        "dynamodb:ListStreams"
    ],
    "Resource": "*"
}


POST_TO_WEBSOCKET_CONNECTION_POLICY = {
    "Effect": "Allow",
    "Action": [
        "execute-api:ManageConnections"
    ],
    "Resource": "arn:*:execute-api:*:*:*/@connections/*"
}



import threading

from typing import Callable, Optional, Type  # noqa
from chalice.local import HTTPServerThread  # noqa


RESTART_REQUEST_RC = 3


class FileWatcher(object):
    """Base class for watching files for changes."""

    def watch_for_file_changes(self, root_dir, callback):
        # type: (str, Callable[[], None]) -> None
        """Recursively watch directory for changes.

        When a changed file is detected, the provided callback
        is immediately invoked and the current scan stops.

        """
        raise NotImplementedError("watch_for_file_changes")


class WorkerProcess(object):
    """Worker that runs the chalice dev server."""
    def __init__(self, http_thread):
        # type: (HTTPServerThread) -> None
        self._http_thread = http_thread
        self._restart_event = threading.Event()

    def main(self, project_dir, timeout=None):
        # type: (str, Optional[int]) -> int
        self._http_thread.start()
        self._start_file_watcher(project_dir)
        if self._restart_event.wait(timeout):
            self._http_thread.shutdown()
            return RESTART_REQUEST_RC
        return 0

    def _start_file_watcher(self, project_dir):
        # type: (str) -> None
        raise NotImplementedError("_start_file_watcher")



import os
import json
import base64
import contextlib
from types import TracebackType

from typing import Optional, Type, Generator, Dict, Any, List  # noqa

from chalice import Chalice  # noqa
from chalice.config import Config
from chalice.local import LocalGateway, LambdaContext, LocalGatewayException
from chalice.cli.factory import CLIFactory


class FunctionNotFoundError(Exception):
    pass


class Client(object):
    def __init__(self, app, stage_name='dev', project_dir='.'):
        # type: (Chalice, str, str) -> None
        self._app = app
        self._project_dir = project_dir
        self._stage_name = stage_name
        self._http_client = None  # type: Optional[TestHTTPClient]
        self._events_client = None  # type: Optional[TestEventsClient]
        self._lambda_client = None  # type: Optional[TestLambdaClient]
        self._chalice_config_obj = None  # type: Optional[Config]
        # We have to be careful about not passing in the CLIFactory
        # because this is a public interface we're exposing and we don't
        # want the CLIFactory to be part of that.
        self._cli_factory = CLIFactory(project_dir)

    @property
    def _chalice_config(self):
        # type: () -> Config
        if self._chalice_config_obj is None:
            try:
                self._chalice_config_obj = self._cli_factory.create_config_obj(
                    chalice_stage_name=self._stage_name)
            except RuntimeError:
                # Being able to load a valid config is not required to use
                # the test client.  If you don't have one that you just won't
                # get any config related data added to your environment.
                self._chalice_config_obj = Config.create()
        return self._chalice_config_obj

    @property
    def http(self):
        # type: () -> TestHTTPClient
        if self._http_client is None:
            self._http_client = TestHTTPClient(self._app, self._chalice_config)
        return self._http_client

    @property
    def lambda_(self):
        # type: () -> TestLambdaClient
        if self._lambda_client is None:
            self._lambda_client = TestLambdaClient(
                self._app, self._chalice_config)
        return self._lambda_client

    @property
    def events(self):
        # type: () -> TestEventsClient
        if self._events_client is None:
            self._events_client = TestEventsClient(self._app)
        return self._events_client

    def __enter__(self):
        # type: () -> Client
        return self

    def __exit__(self,
                 exception_type,   # type: Optional[Type[BaseException]]
                 exception_value,  # type: Optional[BaseException]
                 traceback,        # type: Optional[TracebackType]
                 ):
        # type: (...) -> None
        pass


class BaseClient(object):

    @contextlib.contextmanager
    def _patched_env_vars(self, environment_variables):  # type: ignore
        # re: the "type: ignore".  Mypy is incredibly pedantic that
        # os.environ is of type "_Environ[str]" and it can't be assigned
        # an expression of Dict[str, str].  While that's technically "correct",
        # that's way too much effort to get the typing right on this, so
        # we're just ignoring the type checking for this method.
        original = os.environ
        patched = os.environ.copy()
        patched.update(environment_variables)
        os.environ = patched
        try:
            yield
        finally:
            os.environ = original


class TestHTTPClient(BaseClient):
    def __init__(self, app, config):
        # type: (Chalice, Config) -> None
        self._app = app
        self._config = config
        self._local_gateway = LocalGateway(app, self._config)

    def request(self, method, path, headers=None, body=b''):
        # type: (str, str, Optional[Dict[str, str]], bytes) -> HTTPResponse
        if headers is None:
            headers = {}
        scoped = self._config.scope(self._config.chalice_stage, 'api_handler')
        with self._patched_env_vars(scoped.environment_variables):
            try:
                response = self._local_gateway.handle_request(
                    method=method.upper(), path=path,
                    headers=headers, body=body
                )
            except LocalGatewayException as e:
                return self._error_response(e)
        return HTTPResponse.create_from_dict(response)

    def _error_response(self, e):
        # type: (LocalGatewayException) -> HTTPResponse
        return HTTPResponse(
            headers=e.headers,
            body=e.body if e.body else b'',
            status_code=e.CODE
        )

    def get(self, path, **kwargs):
        # type: (str, **Any) -> HTTPResponse
        return self.request('GET', path, **kwargs)

    def post(self, path, **kwargs):
        # type: (str, **Any) -> HTTPResponse
        return self.request('POST', path, **kwargs)

    def put(self, path, **kwargs):
        # type: (str, **Any) -> HTTPResponse
        return self.request('PUT', path, **kwargs)

    def patch(self, path, **kwargs):
        # type: (str, **Any) -> HTTPResponse
        return self.request('PATCH', path, **kwargs)

    def options(self, path, **kwargs):
        # type: (str, **Any) -> HTTPResponse
        return self.request('OPTIONS', path, **kwargs)

    def delete(self, path, **kwargs):
        # type: (str, **Any) -> HTTPResponse
        return self.request('DELETE', path, **kwargs)

    def head(self, path, **kwargs):
        # type: (str, **Any) -> HTTPResponse
        return self.request('HEAD', path, **kwargs)


class HTTPResponse(object):
    def __init__(self, body, headers, status_code):
        # type: (bytes, Dict[str, str], int) -> None
        self.body = body
        self.headers = headers
        self.status_code = status_code

    @property
    def json_body(self):
        # type: () -> Any
        try:
            return json.loads(self.body)
        except ValueError:
            return None

    @classmethod
    def create_from_dict(cls, response_dict):
        # type: (Dict[str, Any]) -> HTTPResponse
        # Takes the response dict we have to send back to lambda
        # and exposes it as a python object.
        if response_dict.get('isBase64Encoded', False):
            body = base64.b64decode(response_dict['body'])
        else:
            body = response_dict['body'].encode('utf-8')
        combined_headers = response_dict['headers']
        combined_headers.update(response_dict['multiValueHeaders'])
        return cls(
            body=body,
            status_code=response_dict['statusCode'],
            headers=combined_headers,
        )


class TestEventsClient(BaseClient):
    def __init__(self, app):
        # type: (Chalice) -> None
        self._app = app

    def generate_sns_event(self, message, subject=''):
        # type: (str, str) -> Dict[str, Any]
        sns_event = {'Records': [{
            'EventSource': 'aws:sns',
            'EventSubscriptionArn': 'arn:subscription-arn',
            'EventVersion': '1.0',
            'Sns': {
                'Message': message,
                'MessageAttributes': {
                    'AttributeKey': {
                        'Type': 'String',
                        'Value': 'AttributeValue'
                    }
                },
                'MessageId': 'abcdefgh-51e4-5ae2-9964-b296c8d65d1a',
                'Signature': 'signature',
                'SignatureVersion': '1',
                'SigningCertUrl': (
                    'https://sns.us-west-2.amazonaws.com/cert.pem'),
                'Subject': subject,
                'Timestamp': '2018-06-26T19:41:38.695Z',
                'TopicArn': 'arn:aws:sns:us-west-2:12345:TopicName',
                'Type': 'Notification',
                'UnsubscribeUrl': 'https://unsubscribe-url/'
            }
        }]}
        return sns_event

    def generate_s3_event(self, bucket, key, event_name='ObjectCreated:Put'):
        # type: (str, str, str) -> Dict[str, Any]
        s3_event = {
            'Records': [
                {'awsRegion': 'us-west-2',
                 'eventName': event_name,
                 'eventSource': 'aws:s3',
                 'eventTime': '2018-05-22T04:41:23.823Z',
                 'eventVersion': '2.0',
                 'requestParameters': {'sourceIPAddress': '1.1.1.1'},
                 'responseElements': {
                     'x-amz-id-2': 'request-id-2',
                     'x-amz-request-id': 'request-id-1'
                 },
                 's3': {
                     'bucket': {
                         'arn': 'arn:aws:s3:::%s' % bucket,
                         'name': bucket,
                         'ownerIdentity': {
                             'principalId': 'ABCD'
                         }
                     },
                     'configurationId': 'config-id',
                     'object': {
                         'eTag': 'd41d8cd98f00b204e9800998ecf8427e',
                         'key': key,
                         'sequencer': '005B039F73C627CE8B',
                         'size': 0
                     },
                     's3SchemaVersion': '1.0'
                 },
                 'userIdentity': {'principalId': 'AWS:XYZ'}
                 }
            ]
        }
        return s3_event

    def generate_sqs_event(self, message_bodies, queue_name='queue-name'):
        # type: (List[str], str) -> Dict[str, Any]
        records = [{
            'attributes': {
                'ApproximateFirstReceiveTimestamp': '1530576251596',
                'ApproximateReceiveCount': '1',
                'SenderId': 'sender-id',
                'SentTimestamp': '1530576251595'
            },
            'awsRegion': 'us-west-2',
            'body': body,
            'eventSource': 'aws:sqs',
            'eventSourceARN': 'arn:aws:sqs:us-west-2:12345:%s' % queue_name,
            'md5OfBody': '754ac2f7a12df38320e0c5eafd060145',
            'messageAttributes': {},
            'messageId': 'message-id',
            'receiptHandle': 'receipt-handle'
        } for body in message_bodies]
        sqs_event = {'Records': records}
        return sqs_event

    def generate_cw_event(self, source, detail_type,
                          detail, resources, region='us-west-2'):
        # type: (str, str, Dict[str, Any], List[str], str) -> Dict[str, Any]
        event = {
            "version": 0,
            "id": "7bf73129-1428-4cd3-a780-95db273d1602",
            "detail-type": detail_type,
            "source": source,
            "account": "123456789012",
            "time": "2015-11-11T21:29:54Z",
            "region": region,
            "resources": resources,
            "detail": detail,
        }
        return event

    def generate_kinesis_event(self, message_bodies,
                               stream_name='stream-name'):
        # type: (List[bytes], str) -> Dict[str, Any]
        records = [{
            "kinesis": {
                "kinesisSchemaVersion": "1.0",
                "partitionKey": "1",
                "sequenceNumber": "12345",
                "data": base64.b64encode(body).decode('ascii'),
                "approximateArrivalTimestamp": 1545084650.987
            },
            "eventSource": "aws:kinesis",
            "eventVersion": "1.0",
            "eventID": "shardId-000000000006:12345",
            "eventName": "aws:kinesis:record",
            "invokeIdentityArn": "arn:aws:iam::123:role/lambda-role",
            "awsRegion": "us-west-2",
            "eventSourceARN": (
                "arn:aws:kinesis:us-east-2:123:stream/%s" % stream_name
            )
        } for body in message_bodies]
        return {'Records': records}


class TestLambdaClient(BaseClient):
    def __init__(self, app, config):
        # type: (Chalice, Config) -> None
        self._app = app
        self._config = config

    def invoke(self, function_name, payload=None):
        # type: (str, Any) -> InvokeResponse
        if payload is None:
            payload = {}
        scoped = self._config.scope(self._config.chalice_stage, function_name)
        lambda_context = LambdaContext(
            function_name, memory_size=scoped.lambda_memory_size)
        if function_name not in self._app.handler_map:
            raise FunctionNotFoundError(function_name)
        with self._patched_env_vars(scoped.environment_variables):
            response = self._app.handler_map[function_name](
                payload, lambda_context)
        return InvokeResponse(payload=response)


class InvokeResponse(object):
    def __init__(self, payload):
        # type: (Any) -> None
        self.payload = payload



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Resolves regions and endpoints.

This module implements endpoint resolution, including resolving endpoints for a
given service and region and resolving the available endpoints for a service
in a specific AWS partition.
"""
import logging
import re

from botocore.exceptions import NoRegionError

LOG = logging.getLogger(__name__)
DEFAULT_URI_TEMPLATE = '{service}.{region}.{dnsSuffix}'
DEFAULT_SERVICE_DATA = {'endpoints': {}}


class BaseEndpointResolver(object):
    """Resolves regions and endpoints. Must be subclassed."""
    def construct_endpoint(self, service_name, region_name=None):
        """Resolves an endpoint for a service and region combination.

        :type service_name: string
        :param service_name: Name of the service to resolve an endpoint for
            (e.g., s3)

        :type region_name: string
        :param region_name: Region/endpoint name to resolve (e.g., us-east-1)
            if no region is provided, the first found partition-wide endpoint
            will be used if available.

        :rtype: dict
        :return: Returns a dict containing the following keys:
            - partition: (string, required) Resolved partition name
            - endpointName: (string, required) Resolved endpoint name
            - hostname: (string, required) Hostname to use for this endpoint
            - sslCommonName: (string) sslCommonName to use for this endpoint.
            - credentialScope: (dict) Signature version 4 credential scope
              - region: (string) region name override when signing.
              - service: (string) service name override when signing.
            - signatureVersions: (list<string>) A list of possible signature
              versions, including s3, v4, v2, and s3v4
            - protocols: (list<string>) A list of supported protocols
              (e.g., http, https)
            - ...: Other keys may be included as well based on the metadata
        """
        raise NotImplementedError

    def get_available_partitions(self):
        """Lists the partitions available to the endpoint resolver.

        :return: Returns a list of partition names (e.g., ["aws", "aws-cn"]).
        """
        raise NotImplementedError

    def get_available_endpoints(self, service_name, partition_name='aws',
                                allow_non_regional=False):
        """Lists the endpoint names of a particular partition.

        :type service_name: string
        :param service_name: Name of a service to list endpoint for (e.g., s3)

        :type partition_name: string
        :param partition_name: Name of the partition to limit endpoints to.
            (e.g., aws for the public AWS endpoints, aws-cn for AWS China
            endpoints, aws-us-gov for AWS GovCloud (US) Endpoints, etc.

        :type allow_non_regional: bool
        :param allow_non_regional: Set to True to include endpoints that are
             not regional endpoints (e.g., s3-external-1,
             fips-us-gov-west-1, etc).
        :return: Returns a list of endpoint names (e.g., ["us-east-1"]).
        """
        raise NotImplementedError


class EndpointResolver(BaseEndpointResolver):
    """Resolves endpoints based on partition endpoint metadata"""
    def __init__(self, endpoint_data):
        """
        :param endpoint_data: A dict of partition data.
        """
        if 'partitions' not in endpoint_data:
            raise ValueError('Missing "partitions" in endpoint data')
        self._endpoint_data = endpoint_data

    def get_available_partitions(self):
        result = []
        for partition in self._endpoint_data['partitions']:
            result.append(partition['partition'])
        return result

    def get_available_endpoints(self, service_name, partition_name='aws',
                                allow_non_regional=False):
        result = []
        for partition in self._endpoint_data['partitions']:
            if partition['partition'] != partition_name:
                continue
            services = partition['services']
            if service_name not in services:
                continue
            for endpoint_name in services[service_name]['endpoints']:
                if allow_non_regional or endpoint_name in partition['regions']:
                    result.append(endpoint_name)
        return result

    def construct_endpoint(self, service_name, region_name=None, partition_name=None):
        if partition_name is not None:
            valid_partition = None
            for partition in self._endpoint_data['partitions']:
                if partition['partition'] == partition_name:
                    valid_partition = partition

            if valid_partition is not None:
                result = self._endpoint_for_partition(valid_partition, service_name,
                             region_name, True)
                return result
            return None

        # Iterate over each partition until a match is found.
        for partition in self._endpoint_data['partitions']:
            result = self._endpoint_for_partition(
                partition, service_name, region_name)
            if result:
                return result

    def _endpoint_for_partition(self, partition, service_name, region_name,
            force_partition=False):
        # Get the service from the partition, or an empty template.
        service_data = partition['services'].get(
            service_name, DEFAULT_SERVICE_DATA)
        # Use the partition endpoint if no region is supplied.
        if region_name is None:
            if 'partitionEndpoint' in service_data:
                region_name = service_data['partitionEndpoint']
            else:
                raise NoRegionError()
        # Attempt to resolve the exact region for this partition.
        if region_name in service_data['endpoints']:
            return self._resolve(
                partition, service_name, service_data, region_name)
        # Check to see if the endpoint provided is valid for the partition.
        if self._region_match(partition, region_name) or force_partition:
            # Use the partition endpoint if set and not regionalized.
            partition_endpoint = service_data.get('partitionEndpoint')
            is_regionalized = service_data.get('isRegionalized', True)
            if partition_endpoint and not is_regionalized:
                LOG.debug('Using partition endpoint for %s, %s: %s',
                          service_name, region_name, partition_endpoint)
                return self._resolve(
                    partition, service_name, service_data, partition_endpoint)
            LOG.debug('Creating a regex based endpoint for %s, %s',
                      service_name, region_name)
            return self._resolve(
                partition, service_name, service_data, region_name)

    def _region_match(self, partition, region_name):
        if region_name in partition['regions']:
            return True
        if 'regionRegex' in partition:
            return re.compile(partition['regionRegex']).match(region_name)
        return False

    def _resolve(self, partition, service_name, service_data, endpoint_name):
        result = service_data['endpoints'].get(endpoint_name, {})
        result['partition'] = partition['partition']
        result['endpointName'] = endpoint_name
        # Merge in the service defaults then the partition defaults.
        self._merge_keys(service_data.get('defaults', {}), result)
        self._merge_keys(partition.get('defaults', {}), result)
        hostname = result.get('hostname', DEFAULT_URI_TEMPLATE)
        result['hostname'] = self._expand_template(
            partition, result['hostname'], service_name, endpoint_name)
        if 'sslCommonName' in result:
            result['sslCommonName'] = self._expand_template(
                partition, result['sslCommonName'], service_name,
                endpoint_name)
        result['dnsSuffix'] = partition['dnsSuffix']
        return result

    def _merge_keys(self, from_data, result):
        for key in from_data:
            if key not in result:
                result[key] = from_data[key]

    def _expand_template(self, partition, template, service_name,
                         endpoint_name):
        return template.format(
            service=service_name, region=endpoint_name,
            dnsSuffix=partition['dnsSuffix'])


import json
import os
import uuid
from typing import List, Dict, Optional, Any

from aws_cdk import (
    aws_s3_assets as assets,
    cloudformation_include,
    aws_iam as iam,
    aws_lambda as lambda_,
    core as cdk
)

from chalice import api


class Chalice(cdk.Construct):
    """Chalice construct for CDK.

    Packages the application into AWS SAM format and imports the resulting
    template into the construct tree under the provided ``scope``.

    """
    # pylint: disable=redefined-builtin
    # The 'id' parameter name is CDK convention.
    def __init__(self,
                 scope,                       # type: cdk.Construct
                 id,                          # type: str
                 source_dir,                  # type: str
                 stage_config=None,           # type: Optional[Dict[str, Any]]
                 preserve_logical_ids=True,   # type: bool
                 **kwargs                     # type: Dict[str, Any]
                 ):
        # type: (...) -> None
        """Initialize Chalice construct.

        :param str source_dir: Path to Chalice application source code.
        :param dict stage_config: Chalice stage configuration.
            The configuration object should have the same structure as Chalice
            JSON stage configuration.
        :param bool preserve_logical_ids: Whether the resources should have
            the same logical IDs in the resulting CDK template as they did in
            the original CloudFormation template file. If you're vending a
            Construct using cdk-chalice, make sure to pass this as ``False``.
            Note: regardless of whether this option is true or false, the
            :attr:`sam_template`'s ``get_resource`` and related methods always
            uses the original logical ID of the resource/element, as specified
            in the template file.
        :raises `ChaliceError`: Error packaging the Chalice application.
        """
        super(Chalice, self).__init__(scope, id, **kwargs)

        #: (:class:`str`) Path to Chalice application source code.
        self.source_dir = os.path.abspath(source_dir)

        #: (:class:`str`) Chalice stage name.
        #: It is automatically assigned the encompassing CDK ``scope``'s name.
        self.stage_name = scope.to_string()

        #: (:class:`dict`) Chalice stage configuration.
        #: The object has the same structure as Chalice JSON stage
        #: configuration.
        self.stage_config = stage_config

        chalice_out_dir = os.path.join(os.getcwd(), 'chalice.out')
        package_id = uuid.uuid4().hex
        self._sam_package_dir = os.path.join(chalice_out_dir, package_id)

        self._package_app()
        sam_template_filename = self._generate_sam_template_with_assets(
            chalice_out_dir, package_id)

        #: (:class:`aws_cdk.cloudformation_include.CfnInclude`) AWS SAM
        #: template updated with AWS CDK values where applicable. Can be
        #: used to reference, access, and customize resources generated
        #: by `chalice package` commandas CDK native objects.
        self.sam_template = cloudformation_include.CfnInclude(
            self, 'ChaliceApp', template_file=sam_template_filename,
            preserve_logical_ids=preserve_logical_ids)
        self._function_cache = {}  # type: Dict[str, lambda_.IFunction]
        self._role_cache = {}  # type: Dict[str, iam.IRole]

    def _package_app(self):
        # type: () -> None
        api.package_app(
            project_dir=self.source_dir,
            output_dir=self._sam_package_dir,
            stage=self.stage_name,
            chalice_config=self.stage_config,
        )

    def _generate_sam_template_with_assets(self, chalice_out_dir, package_id):
        # type: (str, str) -> str
        deployment_zip_path = os.path.join(
            self._sam_package_dir, 'deployment.zip')
        sam_deployment_asset = assets.Asset(
            self, 'ChaliceAppCode', path=deployment_zip_path)
        sam_template_path = os.path.join(self._sam_package_dir, 'sam.json')
        sam_template_with_assets_path = os.path.join(
            chalice_out_dir, '%s.sam_with_assets.json' % package_id)

        with open(sam_template_path) as sam_template_file:
            sam_template = json.load(sam_template_file)
            for function in self._filter_resources(
                    sam_template, 'AWS::Serverless::Function'):
                function['Properties']['CodeUri'] = {
                    'Bucket': sam_deployment_asset.s3_bucket_name,
                    'Key': sam_deployment_asset.s3_object_key
                }
            managed_layers = self._filter_resources(
                sam_template, 'AWS::Serverless::LayerVersion')
            if len(managed_layers) == 1:
                layer_filename = os.path.join(
                    self._sam_package_dir, 'layer-deployment.zip')
                layer_asset = assets.Asset(
                    self, 'ChaliceManagedLayer', path=layer_filename)
                managed_layers[0]['Properties']['ContentUri'] = {
                    'Bucket': layer_asset.s3_bucket_name,
                    'Key': layer_asset.s3_object_key
                }
        with open(sam_template_with_assets_path, 'w') as f:
            f.write(json.dumps(sam_template, indent=2))
        return sam_template_with_assets_path

    def _filter_resources(self, template, resource_type):
        # type: (Dict[str, Any], str) -> List[Dict[str, Any]]
        return [resource for resource in template['Resources'].values()
                if resource['Type'] == resource_type]

    def get_resource(self, resource_name):
        # type: (str) -> cdk.core.CfnResource
        return self.sam_template.get_resource(resource_name)

    def get_role(self, role_name):
        # type: (str) -> iam.IRole
        if role_name not in self._role_cache:
            cfn_role = self.sam_template.get_resource(role_name)
            # Pylint is incorrectly identifying this as a static method call
            # but it's actually decorated as a @builtins.classmethod method.
            # pylint: disable=no-value-for-parameter
            role = iam.Role.from_role_arn(self, role_name, cfn_role.attr_arn)
            self._role_cache[role_name] = role
        return self._role_cache[role_name]

    def get_function(self, function_name):
        # type: (str) -> lambda_.IFunction
        if function_name not in self._function_cache:
            cfn_lambda = self.sam_template.get_resource(function_name)
            arn_ref = cfn_lambda.get_att('Arn')
            # Pylint is incorrectly identifying this as a static method call
            # but it's actually decorated as a @builtins.classmethod method.
            # pylint: disable=no-value-for-parameter
            function = lambda_.Function.from_function_arn(
                self, function_name, arn_ref.to_string())
            self._function_cache[function_name] = function
        return self._function_cache[function_name]

    def add_environment_variable(self, key, value, function_name):
        # type: (str, str, str) -> None
        cfn_function = self.sam_template.get_resource(function_name)
        cfn_function.add_override(
            'Properties.Environment.Variables.%s' % key, value)



"""Command line interface for chalice.

Contains commands for deploying chalice.

"""
import logging
import os
import platform
import sys
import tempfile
import shutil
import traceback
import functools
import json

import botocore.exceptions
import click
from typing import Dict, Any, Optional, cast  # noqa

from chalice import __version__ as chalice_version
from chalice.app import Chalice  # noqa
from chalice.awsclient import TypedAWSClient
from chalice.awsclient import ReadTimeout
from chalice.cli.factory import CLIFactory
from chalice.cli.factory import NoSuchFunctionError
from chalice.config import Config  # noqa
from chalice.logs import display_logs, LogRetrieveOptions
from chalice.utils import create_zip_file
from chalice.deploy.validate import validate_routes, validate_python_version
from chalice.deploy.validate import ExperimentalFeatureError
from chalice.utils import UI, serialize_to_json
from chalice.constants import DEFAULT_STAGE_NAME
from chalice.local import LocalDevServer  # noqa
from chalice.constants import DEFAULT_HANDLER_NAME
from chalice.invoke import UnhandledLambdaError
from chalice.deploy.swagger import TemplatedSwaggerGenerator
from chalice.deploy.planner import PlanEncoder
from chalice.deploy.appgraph import ApplicationGraphBuilder, GraphPrettyPrint
from chalice.cli import newproj


def _configure_logging(level, format_string=None):
    # type: (int, Optional[str]) -> None
    if format_string is None:
        format_string = "%(asctime)s %(name)s [%(levelname)s] %(message)s"
    logger = logging.getLogger('')
    logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter = logging.Formatter(format_string)
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def get_system_info():
    # type: () -> str
    python_info = "python {}.{}.{}".format(sys.version_info[0],
                                           sys.version_info[1],
                                           sys.version_info[2])
    platform_system = platform.system().lower()
    platform_release = platform.release()
    platform_info = "{} {}".format(platform_system, platform_release)
    return "{}, {}".format(python_info, platform_info)


@click.group()
@click.version_option(version=chalice_version,
                      message='%(prog)s %(version)s, {}'
                      .format(get_system_info()))
@click.option('--project-dir',
              help='The project directory path (absolute or relative).'
                   'Defaults to CWD')
@click.option('--debug/--no-debug',
              default=False,
              help='Print debug logs to stderr.')
@click.pass_context
def cli(ctx, project_dir, debug=False):
    # type: (click.Context, str, bool) -> None
    if project_dir is None:
        project_dir = os.getcwd()
    elif not os.path.isabs(project_dir):
        project_dir = os.path.abspath(project_dir)
    if debug is True:
        _configure_logging(logging.DEBUG)
    _configure_cli_env_vars()
    ctx.obj['project_dir'] = project_dir
    ctx.obj['debug'] = debug
    ctx.obj['factory'] = CLIFactory(project_dir, debug, environ=os.environ)
    os.chdir(project_dir)


def _configure_cli_env_vars():
    # type: () -> None
    # This will set chalice specific env vars so users can detect if
    # we're running a Chalice CLI command.  This is useful if you want
    # conditional behavior only when we're actually running in Lambda
    # in your app.py file.
    os.environ['AWS_CHALICE_CLI_MODE'] = 'true'


@cli.command()
@click.option('--host', default='127.0.0.1')
@click.option('--port', default=8000, type=click.INT)
@click.option('--stage', default=DEFAULT_STAGE_NAME,
              help='Name of the Chalice stage for the local server to use.')
@click.option('--autoreload/--no-autoreload',
              default=True,
              help='Automatically restart server when code changes.')
@click.pass_context
def local(ctx, host='127.0.0.1', port=8000, stage=DEFAULT_STAGE_NAME,
          autoreload=True):
    # type: (click.Context, str, int, str, bool) -> None
    factory = ctx.obj['factory']  # type: CLIFactory
    from chalice.cli import reloader
    # We don't create the server here because that will bind the
    # socket and we only want to do this in the worker process.
    server_factory = functools.partial(
        create_local_server, factory, host, port, stage)
    # When running `chalice local`, a stdout logger is configured
    # so you'll see the same stdout logging as you would when
    # running in lambda.  This is configuring the root logger.
    # The app-specific logger (app.log) will still continue
    # to work.
    logging.basicConfig(
        stream=sys.stdout, level=logging.INFO, format='%(message)s')
    if autoreload:
        project_dir = factory.create_config_obj(
            chalice_stage_name=stage).project_dir
        rc = reloader.run_with_reloader(
            server_factory, os.environ, project_dir)
        # Click doesn't sys.exit() with the RC this function.  The
        # recommended way to do this is to use sys.exit() directly,
        # see: https://github.com/pallets/click/issues/747
        sys.exit(rc)
    run_local_server(factory, host, port, stage)


def create_local_server(factory, host, port, stage):
    # type: (CLIFactory, str, int, str) -> LocalDevServer
    config = factory.create_config_obj(
        chalice_stage_name=stage
    )
    app_obj = config.chalice_app
    # Check that `chalice deploy` would let us deploy these routes, otherwise
    # there is no point in testing locally.
    routes = config.chalice_app.routes
    validate_routes(routes)
    server = factory.create_local_server(app_obj, config, host, port)
    return server


def run_local_server(factory, host, port, stage):
    # type: (CLIFactory, str, int, str) -> None
    server = create_local_server(factory, host, port, stage)
    server.serve_forever()


@cli.command()
@click.option('--autogen-policy/--no-autogen-policy',
              default=None,
              help='Automatically generate IAM policy for app code.')
@click.option('--profile', help='Override profile at deploy time.')
@click.option('--api-gateway-stage',
              help='Name of the API gateway stage to deploy to.')
@click.option('--stage', default=DEFAULT_STAGE_NAME,
              help=('Name of the Chalice stage to deploy to. '
                    'Specifying a new chalice stage will create '
                    'an entirely new set of AWS resources.'))
@click.option('--connection-timeout',
              type=int,
              help=('Overrides the default botocore connection '
                    'timeout.'))
@click.pass_context
def deploy(ctx, autogen_policy, profile, api_gateway_stage, stage,
           connection_timeout):
    # type: (click.Context, Optional[bool], str, str, str, int) -> None
    factory = ctx.obj['factory']  # type: CLIFactory
    factory.profile = profile
    config = factory.create_config_obj(
        chalice_stage_name=stage, autogen_policy=autogen_policy,
        api_gateway_stage=api_gateway_stage,
    )
    session = factory.create_botocore_session(
        connection_timeout=connection_timeout)
    ui = UI()
    d = factory.create_default_deployer(session=session,
                                        config=config,
                                        ui=ui)
    deployed_values = d.deploy(config, chalice_stage_name=stage)
    reporter = factory.create_deployment_reporter(ui=ui)
    reporter.display_report(deployed_values)


@cli.group()
def dev():
    # type: () -> None
    """Development and debugging commands for chalice.

    All the commands under the "chalice dev" namespace are provided
    to help chalice developers introspect the internals of chalice.
    They are also useful for users to better understand the chalice
    deployment process.

    These commands are provided for informational purposes only.
    There is NO guarantee of backwards compatibility for any
    "chalice dev" commands.  Do not rely on the output of these commands.
    These commands allow introspection of chalice internals, and the
    internals of chalice are subject to change as needed.

    """


@dev.command()
@click.option('--autogen-policy/--no-autogen-policy',
              default=None,
              help='Automatically generate IAM policy for app code.')
@click.option('--profile', help='Override profile at deploy time.')
@click.option('--api-gateway-stage',
              help='Name of the API gateway stage to deploy to.')
@click.option('--stage', default=DEFAULT_STAGE_NAME,
              help=('Name of the Chalice stage to deploy to. '
                    'Specifying a new chalice stage will create '
                    'an entirely new set of AWS resources.'))
@click.pass_context
def plan(ctx, autogen_policy, profile, api_gateway_stage, stage):
    # type: (click.Context, Optional[bool], str, str, str) -> None
    """Generate and display deployment plan.

    This command will calculate and pretty print the deployment plan
    without actually executing the plan.  It's primarily used to better
    understand the chalice deployment process.

    """
    factory = ctx.obj['factory']  # type: CLIFactory
    factory.profile = profile
    config = factory.create_config_obj(
        chalice_stage_name=stage, autogen_policy=autogen_policy,
        api_gateway_stage=api_gateway_stage,
    )
    session = factory.create_botocore_session()
    ui = UI()
    d = factory.create_plan_only_deployer(
        session=session, config=config, ui=ui)
    d.deploy(config, chalice_stage_name=stage)


@dev.command()
@click.option('--autogen-policy/--no-autogen-policy',
              default=None,
              help='Automatically generate IAM policy for app code.')
@click.option('--profile', help='Override profile at deploy time.')
@click.option('--api-gateway-stage',
              help='Name of the API gateway stage to deploy to.')
@click.option('--stage', default=DEFAULT_STAGE_NAME,
              help=('Name of the Chalice stage to deploy to. '
                    'Specifying a new chalice stage will create '
                    'an entirely new set of AWS resources.'))
@click.pass_context
def appgraph(ctx, autogen_policy, profile, api_gateway_stage, stage):
    # type: (click.Context, Optional[bool], str, str, str) -> None
    """Generate and display the application graph."""
    factory = ctx.obj['factory']  # type: CLIFactory
    factory.profile = profile
    config = factory.create_config_obj(
        chalice_stage_name=stage, autogen_policy=autogen_policy,
        api_gateway_stage=api_gateway_stage,
    )
    graph_build = ApplicationGraphBuilder()
    graph = graph_build.build(config, stage)
    ui = UI()
    GraphPrettyPrint(ui).display_graph(graph)


@cli.command('invoke')
@click.option('-n', '--name', metavar='NAME', required=True,
              help=('The name of the function to invoke. '
                    'This is the logical name of the function. If the '
                    'function is decorated by app.route use the name '
                    'api_handler instead.'))
@click.option('--profile', metavar='PROFILE',
              help='Override profile at deploy time.')
@click.option('--stage', metavar='STAGE', default=DEFAULT_STAGE_NAME,
              help=('Name of the Chalice stage to deploy to. '
                    'Specifying a new chalice stage will create '
                    'an entirely new set of AWS resources.'))
@click.pass_context
def invoke(ctx, name, profile, stage):
    # type: (click.Context, str, str, str) -> None
    """Invoke the deployed lambda function NAME.

    Reads payload from STDIN.
    """
    factory = ctx.obj['factory']  # type: CLIFactory
    factory.profile = profile

    try:
        invoke_handler = factory.create_lambda_invoke_handler(name, stage)
        payload = factory.create_stdin_reader().read()
        invoke_handler.invoke(payload)
    except NoSuchFunctionError as e:
        err = click.ClickException(
            "could not find a lambda function named %s." % e.name)
        err.exit_code = 2
        raise err
    except botocore.exceptions.ClientError as e:
        error = e.response['Error']
        err = click.ClickException(
            "got '%s' exception back from Lambda\n%s"
            % (error['Code'], error['Message']))
        err.exit_code = 1
        raise err
    except UnhandledLambdaError:
        err = click.ClickException(
            "Unhandled exception in Lambda function, details above.")
        err.exit_code = 1
        raise err
    except ReadTimeout as e:
        err = click.ClickException(e.message)
        err.exit_code = 1
        raise err


@cli.command('delete')
@click.option('--profile', help='Override profile at deploy time.')
@click.option('--stage', default=DEFAULT_STAGE_NAME,
              help='Name of the Chalice stage to delete.')
@click.pass_context
def delete(ctx, profile, stage):
    # type: (click.Context, str, str) -> None
    factory = ctx.obj['factory']  # type: CLIFactory
    factory.profile = profile
    config = factory.create_config_obj(chalice_stage_name=stage)
    session = factory.create_botocore_session()
    d = factory.create_deletion_deployer(session=session, ui=UI())
    d.deploy(config, chalice_stage_name=stage)


@cli.command()
@click.option('--num-entries', default=None, type=int,
              help='Max number of log entries to show.')
@click.option('--include-lambda-messages/--no-include-lambda-messages',
              default=False,
              help='Controls whether or not lambda log messages are included.')
@click.option('--stage', default=DEFAULT_STAGE_NAME,
              help='Name of the Chalice stage to get logs for.')
@click.option('-n', '--name',
              help='The name of the lambda function to retrieve logs from.',
              default=DEFAULT_HANDLER_NAME)
@click.option('-s', '--since',
              help=('Only display logs since the provided time.  If the '
                    '-f/--follow option is specified, then this value will '
                    'default to 10 minutes from the current time.  Otherwise '
                    'by default all log messages are displayed.  This value '
                    'can either be an ISO8601 formatted timestamp or a '
                    'relative time.  For relative times provide a number '
                    'and a single unit.  Units can be "s" for seconds, '
                    '"m" for minutes, "h" for hours, "d" for days, and "w" '
                    'for weeks.  For example "5m" would indicate to display '
                    'logs starting five minutes in the past.'),
              default=None)
@click.option('-f', '--follow/--no-follow',
              default=False,
              help=('Continuously poll for new log messages.  Note that this '
                    'is a best effort attempt, and in certain cases can '
                    'miss log messages.  This option is intended for '
                    'interactive usage only.'))
@click.option('--profile', help='The profile to use for fetching logs.')
@click.pass_context
def logs(ctx, num_entries, include_lambda_messages, stage,
         name, since, follow, profile):
    # type: (click.Context, int, bool, str, str, str, bool, str) -> None
    factory = ctx.obj['factory']  # type: CLIFactory
    factory.profile = profile
    config = factory.create_config_obj(stage, False)
    deployed = config.deployed_resources(stage)
    if name in deployed.resource_names():
        lambda_arn = deployed.resource_values(name)['lambda_arn']
        session = factory.create_botocore_session()
        retriever = factory.create_log_retriever(
            session, lambda_arn, follow)
        options = LogRetrieveOptions.create(
            max_entries=num_entries,
            since=since,
            include_lambda_messages=include_lambda_messages,
        )
        display_logs(retriever, sys.stdout, options)


@cli.command('gen-policy')
@click.option('--filename',
              help='The filename to analyze.  Otherwise app.py is assumed.')
@click.pass_context
def gen_policy(ctx, filename):
    # type: (click.Context, str) -> None
    from chalice import policy
    if filename is None:
        filename = os.path.join(ctx.obj['project_dir'], 'app.py')
    if not os.path.isfile(filename):
        click.echo("App file does not exist: %s" % filename, err=True)
        raise click.Abort()
    with open(filename) as f:
        contents = f.read()
        generated = policy.policy_from_source_code(contents)
        click.echo(serialize_to_json(generated))


@cli.command('new-project')
@click.argument('project_name', required=False)
@click.option('--profile', required=False)
@click.option('-t', '--project-type', required=False, default='legacy')
@click.pass_context
def new_project(ctx, project_name, profile, project_type):
    # type: (click.Context, str, str, str) -> None
    if project_name is None:
        prompter = ctx.obj.get('prompter', newproj.getting_started_prompt)
        answers = prompter()
        project_name = answers['project_name']
        project_type = answers['project_type']
    if os.path.isdir(project_name):
        click.echo("Directory already exists: %s" % project_name, err=True)
        raise click.Abort()
    newproj.create_new_project_skeleton(
        project_name, project_type=project_type)
    validate_python_version(Config.create())
    click.echo("Your project has been generated in ./%s" % project_name)


@cli.command('url')
@click.option('--stage', default=DEFAULT_STAGE_NAME,
              help='Name of the Chalice stage to get the deployed URL for.')
@click.pass_context
def url(ctx, stage):
    # type: (click.Context, str) -> None
    factory = ctx.obj['factory']  # type: CLIFactory
    config = factory.create_config_obj(stage)
    deployed = config.deployed_resources(stage)
    if deployed is not None and 'rest_api' in deployed.resource_names():
        click.echo(deployed.resource_values('rest_api')['rest_api_url'])
    else:
        e = click.ClickException(
            "Could not find a record of a Rest API in chalice stage: '%s'"
            % stage)
        e.exit_code = 2
        raise e


@cli.command('generate-sdk')
@click.option('--sdk-type', default='javascript',
              type=click.Choice(['javascript']))
@click.option('--stage', default=DEFAULT_STAGE_NAME,
              help='Name of the Chalice stage to generate an SDK for.')
@click.argument('outdir')
@click.pass_context
def generate_sdk(ctx, sdk_type, stage, outdir):
    # type: (click.Context, str, str, str) -> None
    factory = ctx.obj['factory']  # type: CLIFactory
    config = factory.create_config_obj(stage)
    session = factory.create_botocore_session()
    client = TypedAWSClient(session)
    deployed = config.deployed_resources(stage)
    if deployed is not None and 'rest_api' in deployed.resource_names():
        rest_api_id = deployed.resource_values('rest_api')['rest_api_id']
        api_gateway_stage = config.api_gateway_stage
        client.download_sdk(rest_api_id, outdir,
                            api_gateway_stage=api_gateway_stage,
                            sdk_type=sdk_type)
    else:
        click.echo("Could not find API ID, has this application "
                   "been deployed?", err=True)
        raise click.Abort()


@cli.command('generate-models')
@click.option('--stage', default=DEFAULT_STAGE_NAME,
              help="Chalice Stage for which to generate models.")
@click.pass_context
def generate_models(ctx, stage):
    # type: (click.Context, str) -> None
    """Generate a model from Chalice routes.

    Currently only supports generating Swagger 2.0 models.
    """
    factory = ctx.obj['factory']  # type: CLIFactory
    config = factory.create_config_obj(stage)
    if not config.chalice_app.routes:
        click.echo('No REST API found to generate model from.')
        raise click.Abort()
    swagger_generator = TemplatedSwaggerGenerator()
    model = swagger_generator.generate_swagger(
        config.chalice_app,
    )
    ui = UI()
    ui.write(json.dumps(model, indent=4, cls=PlanEncoder))
    ui.write('\n')


@cli.command('package')
@click.option('--pkg-format', default='cloudformation',
              help=('Specify the provisioning engine to use for '
                    'template output. Chalice supports both '
                    'CloudFormation and Terraform. Default '
                    'is CloudFormation.'),
              type=click.Choice(['cloudformation', 'terraform']))
@click.option('--stage', default=DEFAULT_STAGE_NAME,
              help="Chalice Stage to package.")
@click.option('--single-file', is_flag=True,
              default=False,
              help=("Create a single packaged file. "
                    "By default, the 'out' argument "
                    "specifies a directory in which the "
                    "package assets will be placed.  If "
                    "this argument is specified, a single "
                    "zip file will be created instead. CloudFormation Only."))
@click.option('--merge-template',
              help=('Specify a JSON or YAML template to be merged '
                    'into the generated template. This is useful '
                    'for adding resources to a Chalice template or '
                    'modify values in the template. CloudFormation Only.'))
@click.option('--template-format', default='json',
              type=click.Choice(['json', 'yaml'], case_sensitive=False),
              help=('Specify if the generated template should be serialized '
                    'as either JSON or YAML.  CloudFormation only.'))
@click.option('--profile', help='Override profile at packaging time.')
@click.argument('out')
@click.pass_context
def package(ctx, single_file, stage, merge_template,
            out, pkg_format, template_format, profile):
    # type: (click.Context, bool, str, str, str, str, str, str) -> None
    factory = ctx.obj['factory']  # type: CLIFactory
    factory.profile = profile
    config = factory.create_config_obj(stage)
    options = factory.create_package_options()
    packager = factory.create_app_packager(config, options,
                                           pkg_format,
                                           template_format,
                                           merge_template)
    if pkg_format == 'terraform' and (merge_template or
                                      single_file or
                                      template_format != 'json'):
        # I don't see any reason we couldn't support --single-file for
        # terraform if we wanted to.
        click.echo((
            "Terraform format does not support "
            "--merge-template, --single-file, or --template-format"))
        raise click.Abort()

    if single_file:
        dirname = tempfile.mkdtemp()
        try:
            packager.package_app(config, dirname, stage)
            create_zip_file(source_dir=dirname, outfile=out)
        finally:
            shutil.rmtree(dirname)
    else:
        packager.package_app(config, out, stage)


@cli.command('generate-pipeline')
@click.option('--pipeline-version',
              default='v1',
              type=click.Choice(['v1', 'v2']),
              help='Which version of the pipeline template to generate.')
@click.option('-i', '--codebuild-image',
              help=("Specify default codebuild image to use.  "
                    "This option must be provided when using a python "
                    "version besides 2.7."))
@click.option('-s', '--source', default='codecommit',
              type=click.Choice(['codecommit', 'github']),
              help=("Specify the input source.  The default value of "
                    "'codecommit' will create a CodeCommit repository "
                    "for you.  The 'github' value allows you to "
                    "reference an existing GitHub repository."))
@click.option('-b', '--buildspec-file',
              help=("Specify path for buildspec.yml file. "
                    "By default, the build steps are included in the "
                    "generated cloudformation template.  If this option "
                    "is provided, a buildspec.yml will be generated "
                    "as a separate file and not included in the cfn "
                    "template.  This allows you to make changes to how "
                    "the project is built without having to redeploy "
                    "a CloudFormation template. This file should be "
                    "named 'buildspec.yml' and placed in the root "
                    "directory of your app."))
@click.argument('filename')
@click.pass_context
def generate_pipeline(ctx, pipeline_version, codebuild_image, source,
                      buildspec_file, filename):
    # type: (click.Context, str, str, str, str, str) -> None
    """Generate a cloudformation template for a starter CD pipeline.

    This command will write a starter cloudformation template to
    the filename you provide.  It contains a CodeCommit repo,
    a CodeBuild stage for packaging your chalice app, and a
    CodePipeline stage to deploy your application using cloudformation.

    You can use any AWS SDK or the AWS CLI to deploy this stack.
    Here's an example using the AWS CLI:

        \b
        $ chalice generate-pipeline pipeline.json
        $ aws cloudformation deploy --stack-name mystack \b
            --template-file pipeline.json --capabilities CAPABILITY_IAM
    """
    from chalice import pipeline
    factory = ctx.obj['factory']  # type: CLIFactory
    config = factory.create_config_obj()
    p = cast(pipeline.BasePipelineTemplate, None)
    if pipeline_version == 'v1':
        p = pipeline.CreatePipelineTemplateLegacy()
    else:
        p = pipeline.CreatePipelineTemplateV2()
    params = pipeline.PipelineParameters(
        app_name=config.app_name,
        lambda_python_version=config.lambda_python_version,
        codebuild_image=codebuild_image,
        code_source=source,
        pipeline_version=pipeline_version,
    )
    output = p.create_template(params)
    if buildspec_file:
        extractor = pipeline.BuildSpecExtractor()
        buildspec_contents = extractor.extract_buildspec(output)
        with open(buildspec_file, 'w') as f:
            f.write(buildspec_contents)
    with open(filename, 'w') as f:
        f.write(serialize_to_json(output))


def main():
    # type: () -> int
    # click's dynamic attrs will allow us to pass through
    # 'obj' via the context object, so we're ignoring
    # these error messages from pylint because we know it's ok.
    # pylint: disable=unexpected-keyword-arg,no-value-for-parameter
    try:
        return cli(obj={})
    except botocore.exceptions.NoRegionError:
        click.echo("No region configured. "
                   "Either export the AWS_DEFAULT_REGION "
                   "environment variable or set the "
                   "region value in our ~/.aws/config file.", err=True)
        return 2
    except ExperimentalFeatureError as e:
        click.echo(str(e))
        return 2
    except Exception:
        click.echo(traceback.format_exc(), err=True)
        return 2



"""Module for inspecting chalice logs.

This module provides APIs for searching, interacting
with the logs generated by AWS Lambda.

"""
import time
from datetime import datetime, timedelta
from collections import defaultdict

from typing import Any, Optional, Iterator, Dict, IO, Callable, Set  # noqa
from botocore.session import Session  # noqa
from attr import attrs, attrib

from chalice.awsclient import TypedAWSClient, CWLogEvent  # noqa
from chalice.utils import TimestampConverter


@attrs
class LogRetrieveOptions(object):
    max_entries = attrib(default=None)               # type: Optional[int]
    start_time = attrib(default=None)                # type: Optional[datetime]
    include_lambda_messages = attrib(default=False)  # type: bool

    @classmethod
    def create(cls, follow=False, since=None, **kwargs):
        # type: (bool, Optional[str], **Any) -> LogRetrieveOptions
        start_time = None
        if since is not None:
            start_time = TimestampConverter().timestamp_to_datetime(since)
        elif follow:
            # If they've specified --follow with no start time, we'll default
            # to 10 minutes prior to now.  This is a backwards compat
            # quirk where we can't change the default for `chalice logs`, but
            # for the --follow mode, we can default to only showing 10 minutes
            # worth of logs before we start polling for new logs.  In a
            # future major version we can just change the default for all log
            # retrieval to 10 minutes.
            start_time = datetime.utcnow() - timedelta(minutes=10)
        kwargs['start_time'] = start_time
        return cls(**kwargs)


def display_logs(retriever, stream, retrieve_options):
    # type: (LogRetriever, IO[str], LogRetrieveOptions) -> None
    events = retriever.retrieve_logs(retrieve_options)
    for event in events:
        stream.write('%s %s %s\n' % (
            event['timestamp'],
            event['logShortId'],
            event['message'].strip()))


class LogRetriever(object):
    def __init__(self, log_event_generator, log_group_name):
        # type: (BaseLogEventGenerator, str) -> None
        self._log_event_generator = log_event_generator
        self._log_group_name = log_group_name

    @classmethod
    def create_from_lambda_arn(cls, log_event_generator, lambda_arn):
        # type: (BaseLogEventGenerator, str) -> LogRetriever
        """Create a LogRetriever from a client and lambda arn.

        :type client: botocore.client.Logs
        :param client: A ``logs`` client.

        :type lambda_arn: str
        :param lambda_arn: The ARN of the lambda function.

        :return: An instance of ``LogRetriever``.

        """
        lambda_name = lambda_arn.split(':')[6]
        log_group_name = '/aws/lambda/%s' % lambda_name
        return cls(log_event_generator, log_group_name)

    def _is_lambda_message(self, event):
        # type: (CWLogEvent) -> bool
        # Lambda will also inject log messages into your log streams.
        # They look like:
        # START RequestId: guid Version: $LATEST
        # END RequestId: guid
        # REPORT RequestId: guid    Duration: 0.35 ms   Billed Duration: ...

        # By default, these message are included in retrieve_logs().
        # But you can also request that retrieve_logs() filter out
        # these message so that we only include log messages generated
        # by your chalice app.
        msg = event['message'].strip()
        return msg.startswith(('START RequestId',
                               'END RequestId',
                               'REPORT RequestId'))

    def retrieve_logs(self, retrieve_options):
        # type: (LogRetrieveOptions) -> Iterator[CWLogEvent]
        """Retrieve logs from a log group.

        :type include_lambda_messages: boolean
        :param include_lambda_messages: Include logs generated by the AWS
            Lambda service.  If this value is False, only chalice logs will be
            included.

        :type max_entries: int
        :param max_entries: Maximum number of log messages to include.

        :rtype: iterator
        :return: An iterator that yields event dicts.  Each event
            dict has these keys:

            * logStreamName -> (string) The name of the log stream.
            * timestamp -> (datetime.datetime) - The timestamp for the msg.
            * message -> (string) The data contained in the log event.
            * ingestionTime -> (datetime.datetime) Ingestion time of event.
            * eventId -> (string) A unique identifier for this event.
            * logShortId -> (string) Short identifier for logStreamName.

        """
        max_entries = retrieve_options.max_entries
        shown = 0
        events = self._log_event_generator.iter_log_events(
            self._log_group_name, retrieve_options)
        for event in events:
            if not retrieve_options.include_lambda_messages and \
                    self._is_lambda_message(event):
                continue
            # logStreamName is: '2016/07/05/[id]hash'
            # We want to extract the hash portion and
            # provide a short identifier.
            identifier = event['logStreamName']
            if ']' in identifier:
                index = identifier.find(']')
                identifier = identifier[index + 1:index + 7]
            event['logShortId'] = identifier
            yield event
            shown += 1
            if max_entries is not None and shown >= max_entries:
                return


class BaseLogEventGenerator(object):
    def __init__(self, client):
        # type: (TypedAWSClient) -> None
        self._client = client

    def iter_log_events(self, log_group_name, options):
        # type: (str, LogRetrieveOptions) -> Iterator[CWLogEvent]
        pass


class LogEventGenerator(BaseLogEventGenerator):
    def iter_log_events(self, log_group_name, options):
        # type: (str, LogRetrieveOptions) -> Iterator[CWLogEvent]
        logs = self._client.iter_log_events(log_group_name=log_group_name,
                                            start_time=options.start_time)
        for log_message in logs:
            yield log_message


class FollowLogEventGenerator(BaseLogEventGenerator):

    _POLL_TIME = 5

    def __init__(self, client, sleep=time.sleep, poll_time=_POLL_TIME):
        # type: (TypedAWSClient, Callable[[int], None], int) -> None
        self._client = client
        self._sleep = sleep
        self._event_id_cache = defaultdict(
            set)  # type: Dict[datetime, Set[str]]
        self._poll_time = poll_time

    def iter_log_events(self, log_group_name, options):
        # type: (str, LogRetrieveOptions) -> Iterator[CWLogEvent]
        start_time = options.start_time
        try:
            for event in self._loop_on_filter_log_events(log_group_name,
                                                         start_time):
                yield event
        except KeyboardInterrupt:
            return

    def _loop_on_filter_log_events(self, log_group_name, start_time):
        # type: (str, Optional[datetime]) -> Iterator[CWLogEvent]
        self._event_id_cache.clear()
        kwargs = {
            'log_group_name': log_group_name,
            'start_time': start_time,
        }  # type: Dict[str, Any]
        while True:
            response = self._client.filter_log_events(**kwargs)
            for event in response['events']:
                if not self._in_cache(event):
                    self._add_to_cache(event)
                    yield event
            if 'nextToken' in response:
                # If there's more pages we go through the normal pagination
                # logic of following nextTokens.
                kwargs['next_token'] = response['nextToken']
            else:
                kwargs.pop('next_token', None)
                if self._event_id_cache:
                    # However, if there's no nextToken it means we've iterated
                    # through all the existing log events.  We now need to
                    # start polling for new events.  To do this, we need
                    # to use start time and not nextToken, because once
                    # we've exhausted the iteration we won't see new events
                    # from any log streams we've already iterated through.
                    # Querying only based on start time ensures we see new
                    # messages from all new streams.
                    # This isn't going to be perfect.  It's possible that
                    # we miss a gap between when we finished searching through
                    # a log stream and the new start time we're going to use
                    # to start polling, especially at high rates of log
                    # generation.
                    most_recent_start_time = max(self._event_id_cache)
                    kwargs['start_time'] = most_recent_start_time
                    self._prune_old_cache_entries(most_recent_start_time)
                    self._sleep(self._poll_time)

    def _in_cache(self, event):
        # type: (CWLogEvent) -> bool
        return event['eventId'] in self._event_id_cache[event['timestamp']]

    def _add_to_cache(self, event):
        # type: (CWLogEvent) -> None
        self._event_id_cache[event['timestamp']].add(event['eventId'])

    def _prune_old_cache_entries(self, timestamp):
        # type: (datetime) -> None
        surviving_entries = self._event_id_cache[timestamp]
        self._event_id_cache.clear()
        self._event_id_cache[timestamp] = surviving_entries



"""Abstraction for invoking a lambda function."""
import json

from typing import Any, Optional, Dict, List, Union, Tuple  # noqa

from chalice.config import DeployedResources  # noqa
from chalice.awsclient import TypedAWSClient  # noqa
from chalice.utils import UI  # noqa
from chalice.compat import StringIO


OptBytes = Optional[bytes]
_ERROR_KEY = 'FunctionError'
_ERROR_VALUE = 'Unhandled'


def _response_is_error(response):
    # type: (Dict[str, Any]) -> bool
    return response.get(_ERROR_KEY) == _ERROR_VALUE


class UnhandledLambdaError(Exception):
    pass


class LambdaInvokeHandler(object):
    """Handler class to coordinate making an invoke call to lambda.

    This class takes a LambdaInvoker, a LambdaResponseFormatter, and a UI
    object in order to make an invoke call against lambda, format the response
    and render it to the UI.
    """
    def __init__(self, invoker, formatter, ui):
        # type: (LambdaInvoker, LambdaResponseFormatter, UI) -> None
        self._invoker = invoker
        self._formatter = formatter
        self._ui = ui

    def invoke(self, payload=None):
        # type: (OptBytes) -> None
        response = self._invoker.invoke(payload)
        formatted_response = self._formatter.format_response(response)
        if _response_is_error(response):
            self._ui.error(formatted_response)
            raise UnhandledLambdaError()
        self._ui.write(formatted_response)


class LambdaInvoker(object):
    def __init__(self, lambda_arn, client):
        # type: (str, TypedAWSClient) -> None
        self._lambda_arn = lambda_arn
        self._client = client

    def invoke(self, payload=None):
        # type: (OptBytes) -> Dict[str, Any]
        return self._client.invoke_function(
            self._lambda_arn,
            payload=payload
        )


class LambdaResponseFormatter(object):
    _PAYLOAD_KEY = 'Payload'

    _TRACEBACK_HEADING = 'Traceback (most recent call last):\n'

    def format_response(self, response):
        # type: (Dict[str, Any]) -> str
        formatted = StringIO()
        payload = response[self._PAYLOAD_KEY].read()
        if _response_is_error(response):
            self._format_error(formatted, payload)
        else:
            self._format_success(formatted, payload)
        return str(formatted.getvalue())

    def _format_error(self, formatted, payload):
        # type: (StringIO, bytes) -> None
        loaded_error = json.loads(payload)
        error_message = loaded_error['errorMessage']
        error_type = loaded_error.get('errorType')
        stack_trace = loaded_error.get('stackTrace')

        if stack_trace is not None:
            self._format_stacktrace(formatted, stack_trace)

        if error_type is not None:
            formatted.write('{}: {}\n'.format(error_type, error_message))
        else:
            formatted.write('{}\n'.format(error_message))

    def _format_stacktrace(self, formatted, stack_trace):
        # type: (StringIO, List[List[Union[str, int]]]) -> None
        formatted.write(self._TRACEBACK_HEADING)
        for frame in stack_trace:
            self._format_frame(formatted, frame)

    def _format_frame(self, formatted, frame):
        # type: (StringIO, Union[str, List[Union[str, int]]]) -> None
        if isinstance(frame, list):
            # If the output is a list, it came from a 4-tuple as a result of
            # an extract_tb call. This is the behavior up to and including
            # python 3.6.
            path, lineno, function, code = frame
            formatted.write(
                '  File "{}", line {}, in {}\n'.format(path, lineno, function))
            formatted.write(
                '    {}\n'.format(code))
        else:
            # If it is not a list, its a string. This is because the 4-tuple
            # was replaced with a FrameSummary object which is serialized as
            # a string by Lambda. In this case we can just print it directly.
            formatted.write(frame)

    def _format_success(self, formatted, payload):
        # type: (StringIO, bytes) -> None
        formatted.write('{}\n'.format(str(payload.decode('utf-8'))))



import socket
import six
import os

from typing import Dict, Any  # noqa
from urllib.parse import urlparse, parse_qs

from six import StringIO


STRING_TYPES = six.string_types


def pip_import_string():
    # type: () -> str
    import pip
    pip_major_version = int(pip.__version__.split('.')[0])
    pip_minor_version = int(pip.__version__.split('.')[1])
    pip_major_minor = (pip_major_version, pip_minor_version)
    # Pip moved its internals to an _internal module in version 10.
    # In order to be compatible with version 9 which has it at at the
    # top level we need to figure out the correct import path here.
    if (9, 0) <= pip_major_minor < (10, 0):
        return 'from pip import main'
    elif (10, 0) <= pip_major_minor < (19, 3):
        # Pip changed their import structure again in 19.3
        # https://github.com/pypa/pip/commit/09fd200
        return 'from pip._internal import main'
    elif (19, 3) <= pip_major_minor < (20, 0):
        return 'from pip._internal.main import main'
    elif (20, 0) <= pip_major_minor < (22, 0):
        # More changes! https://github.com/pypa/pip/issues/7498
        return 'from pip._internal.cli.main import main'
    raise RuntimeError("Unknown import string for pip version: %s"
                       % str(pip_major_minor))


if os.name == 'nt':
    # windows
    # This is the actual patch used on windows to prevent distutils from
    # compiling C extensions. The msvc compiler base class has its compile
    # method overridden to raise a CompileError. This can be caught by
    # setup.py code which can then fallback to making a pure python
    # package if possible.
    # We need mypy to ignore these since they are never actually called from
    # within our process they do not need to be a part of our typechecking
    # pass.
    def prevent_msvc_compiling_patch():  # type: ignore
        import distutils
        import distutils._msvccompiler
        import distutils.msvc9compiler
        import distutils.msvccompiler

        from distutils.errors import CompileError

        def raise_compile_error(*args, **kwargs):  # type: ignore
            raise CompileError('Chalice blocked C extension compiling.')
        distutils._msvccompiler.MSVCCompiler.compile = raise_compile_error
        distutils.msvc9compiler.MSVCCompiler.compile = raise_compile_error
        distutils.msvccompiler.MSVCCompiler.compile = raise_compile_error

    # This is the setuptools shim used to execute setup.py by pip.
    # Lines 2 and 3 have been added to call the above function
    # `prevent_msvc_compiling_patch` and extra escapes have been added on line
    # 5 because it is passed through another layer of string parsing before it
    # is executed.
    _SETUPTOOLS_SHIM = (
        r"import setuptools, tokenize;__file__=%r;"
        r"from chalice.compat import prevent_msvc_compiling_patch;"
        r"prevent_msvc_compiling_patch();"
        r"f=getattr(tokenize, 'open', open)(__file__);"
        r"code=f.read().replace('\\r\\n', '\\n');"
        r"f.close();"
        r"exec(compile(code, __file__, 'exec'))"
    )

    # On windows the C compiling story is much more complex than on posix as
    # there are many different C compilers that setuptools and disutils will
    # try and find using a combination of known filepaths, registry entries,
    # and environment variables. Since there is no simple environment variable
    # we can replace when starting the subprocess that builds the package;
    # we need to apply a patch at runtime to prevent pip/setuptools/distutils
    # from being able to build C extensions.
    # Patching out every possible technique for finding each compiler would
    # be a losing game of whack-a-mole. In addition we need to apply a patch
    # two layers down through subprocess calls, specifically:
    #  * Chalice creates a subprocess of `pip wheel ...` to build sdists
    #    into wheel files.
    #  * Pip creates another python subprocess to call the setup.py file in
    #    the sdist. Before doing so it applies the above shim to make the
    #    setup file compatible with setuptools. This shim layer also reads
    #    and executes the code in the setup.py.
    #  * Setuptools (which will have been executed by the shim) will
    #    eventually call distutils to do the heavy lifting for C compiling.
    #
    # Our patch needs to affect the bottom level here (distutils) and patch
    # it out to prevent it from compiling C in a graceful way that results in
    # falling back to building a purepython library if possible.
    # The below line will be injected just before the `pip wheel ...` portion
    # of the subprocess that Chalice starts. This replaces the
    # SETUPTOOLS_SHIM that pip normally uses with the one defined above.
    # When pip goes to run its subprocess for executing setup.py it will
    # inject _SETUPTOOLS_SHIM rather than the usual SETUPTOOLS_SHIM in pip.
    # This lets us apply our patches in the same process that will compile
    # the c extensions before the setup.py file has been executed.
    # The actual patches used are decribed in the comment above
    # _SETUPTOOLS_SHIM.
    pip_no_compile_c_shim = (
        'import pip;'
        'pip.wheel.SETUPTOOLS_SHIM = """%s""";'
    ) % _SETUPTOOLS_SHIM
    pip_no_compile_c_env_vars = {}  # type: Dict[str, Any]
else:
    # posix
    # On posix systems setuptools/distutils uses the CC env variable to
    # locate a C compiler for building C extensions. All we need to do is set
    # it to /var/false, and the module building process will fail to build.
    # C extensions, and any fallback processes in place to build a pure python
    # package will be kicked off.
    # No need to monkey patch the process.
    pip_no_compile_c_shim = ''
    pip_no_compile_c_env_vars = {
        'CC': '/var/false'
    }


def is_broken_pipe_error(error):
    # type: (Exception) -> bool
    return isinstance(error, BrokenPipeError)  # noqa



from typing import (
    Dict,
    List,
    Any,
    Callable,
    Union,
    Optional,
    Set,
    Type,
    Mapping,
    MutableMapping,
    Sequence,
    Iterator,
)
import logging
import datetime
from chalice.local import LambdaContext

__version__ = ... # type: str

class ChaliceError(Exception): ...
class ChaliceUnhandledError(ChaliceError): ...
class WebsocketDisconnectedError(Exception): ...
class ChaliceViewError(ChaliceError):
    __name__ = ... # type: str
    STATUS_CODE = ... # type: int
class BadRequestError(ChaliceViewError): ...
class UnauthorizedError(ChaliceViewError): ...
class ForbiddenError(ChaliceViewError): ...
class NotFoundError(ChaliceViewError): ...
class ConflictError(ChaliceViewError): ...
class UnprocessableEntityError(ChaliceViewError): ...
class TooManyRequestsError(ChaliceViewError): ...
class RequestTimeoutError(ChaliceViewError): ...


ALL_ERRORS = ... # type: List[ChaliceViewError]
_GET_RESPONSE = Callable[[Any], Any]
_MIDDLEWARE_FUNC = Callable[[Any, _GET_RESPONSE], Any]


class Authorizer:
    name = ... # type: str
    scopes = ... # type: List[str]
    def to_swagger(self) -> Dict[str, Any]: ...
    def with_scopes(self, scopes: List[str]) -> Authorizer: ... #pylint: disable=undefined-variable


class CognitoUserPoolAuthorizer(Authorizer):
    def __init__(
        self,
        name: str,
        provider_arns: List[str],
        header: Optional[str],
        scopes: Optional[List]
    ) -> None: ...

class IAMAuthorizer(Authorizer): ...

class CustomAuthorizer(Authorizer):

    def __init__(self, name: str=...,
                 authorizer_uri: str=...,
                 ttl_seconds: int=...,
                 header: str=...,
                 invoke_role_arn: Optional[str]=...,
                 scopes: Optional[List[str]]=...) -> None: ...


class CORSConfig:
    _REQUIRED_HEADERS = ... # type: List[str]
    allow_origin = ... # type: str
    allow_headers = ... # type: Optional[Sequence[str]]
    get_access_control_headers = ... # type: Callable[..., Dict[str, str]]

    def __init__(self, allow_origin: str=...,
                 allow_headers: Optional[Sequence[str]]=...,
                 expose_headers: Optional[Sequence[str]]=...,
                 max_age: Optional[int]=...,
                 allow_credentials: Optional[bool]=...) -> None: ...

    def __eq__(self, other: object) -> bool: ...


class Request:
    query_params = ... # type: Optional[Dict[str, str]]
    headers = ... # type: CaseInsensitiveMapping
    uri_params = ... # type: Optional[Dict[str, str]]
    method = ... # type: str
    body = ... # type: Any
    base64_body = ... # type: str
    context = ... # type: Dict[str, Any]
    stage_vars = ... # type: Optional[Dict[str, str]]
    json_body = ... # type: Any
    path = ... # type: str
    _json_body = ... # type: Optional[Any]
    raw_body = ... # type: Optional[Any]

    def __init__(
        self, event_dict: Dict[str, Any], lambda_context: Optional[Any]
    ) -> None: ...
    def to_dict(self) -> Dict[Any, Any]: ...


class Response:
    headers = ... # type: Dict[str, Union[str, List[str]]]
    body = ...  # type: Any
    status_code = ... # type: int

    def __init__(self,
                 body: Any,
                 headers: Optional[Dict[str, Union[str, List[str]]]]=...,
                 status_code: int=...) -> None: ...

    def to_dict(self,
                binary_types: Optional[List[str]]=...) -> Dict[str, Any]: ...


class RouteEntry(object):
    view_function = ... # type: Callable[..., Any]
    view_name = ... # type: str
    method = ... # type: str
    uri_pattern = ... # type: str
    authorizer_name = ... # type: str
    authorizer = ... # type: Optional[Authorizer]
    api_key_required = ... # type: bool
    content_types = ... # type: List[str]
    view_args = ... # type: List[str]
    cors = ... # type: CORSConfig

    def __init__(self,
                 view_function: Callable[..., Any],
                 view_name: str,
                 path: str,
                 method: str,
                 api_key_required: Optional[bool]=...,
                 content_types: Optional[List[str]]=...,
                 authorizer: Optional[Union[Authorizer,
                                            ChaliceAuthorizer]]=...,
                 cors: Union[bool, CORSConfig]=...) -> None: ...

    def _parse_view_args(self) -> List[str]: ...

    def __eq__(self, other: object) -> bool: ...


class APIGateway(object):
    cors = ... # type: Union[bool, CORSConfig]
    binary_types = ... # type: List[str]
    default_binary_types = ... # type: List[str]


class WebsocketAPI(object):
    session = ... # type: Optional[Any]

    def configure(self,
                  domain_name: str,
                  stage: str) -> None: ...

    def send(self,
             connection_id: str,
             message: str) -> None: ...


class DecoratorAPI(object):
    def register_middleware(self,
                            func: _MIDDLEWARE_FUNC,
                            event_type: str=...) -> None: ...

    def middleware(self, event_type: str=...) -> Callable[..., Any]: ...

    def authorizer(self,
                   ttl_seconds: Optional[int]=...,
                   execution_role: Optional[str]=...,
                   name: Optional[str]=...,
                   header: Optional[str]=...) -> Callable[..., Any]: ...

    def on_s3_event(self,
                    bucket: str,
                    events: Optional[List[str]]=...,
                    prefix: Optional[str]=...,
                    suffix: Optional[str]=...,
                    name: Optional[str]=...) -> Callable[..., Any]: ...

    def on_sns_message(self,
                      topic: str,
                      name: Optional[str]=...) -> Callable[..., Any]: ...

    def on_sqs_message(self,
                       queue: str,
                       batch_size: int=...,
                       name: Optional[str]=...) -> Callable[..., Any]: ...

    def on_cw_event(self,
                    event_pattern: Dict[str, Any],
                    name: Optional[str]=None) -> Callable[..., Any]: ...

    def on_kinesis_record(self,
                          stream: str,
                          batch_size: int=...,
                          startition_position: str=...,
                          name: Optional[str]=...) -> Callable[..., Any]: ...

    def on_dynamodb_record(self,
                           stream_arn: str,
                           batch_size: int=...,
                           startition_position: str=...,
                           name: Optional[str]=...) -> Callable[..., Any]: ...

    def schedule(self,
                 expression: Union[str, Cron, Rate],
                 name: Optional[str]=...,
                 description: Optional[str]=...) -> Callable[..., Any]: ...

    def route(self, path: str, **kwargs: Any) -> Callable[..., Any]: ...

    def lambda_function(self, name: Optional[str]=...) -> Callable[..., Any]: ...

    def on_ws_connect(self, name: Optional[str] = ...) -> Callable[..., Any]: ...

    def on_ws_disconnect(self, name: Optional[str] = ...) -> Callable[..., Any]: ...

    def on_ws_message(self, name: Optional[str] = ...) -> Callable[..., Any]: ...


class Chalice(DecoratorAPI):
    app_name = ... # type: str
    api = ... # type: APIGateway
    routes = ... # type: Dict[str, Dict[str, RouteEntry]]
    websocket_api = ... # type: WebsocketAPI
    websocket_handlers = ... # type: Dict[str, Any]
    current_request = ... # type: Request
    lambda_context = ... # type: LambdaContext
    debug = ... # type: bool
    configure_logs = ... # type: bool
    log = ... # type: logging.Logger
    authorizers = ... # type: Dict[str, Dict[str, Any]]
    builtin_auth_handlers = ... # type: List[BuiltinAuthConfig]
    event_sources = ... # type: List[Type[BaseEventSourceConfig]]
    pure_lambda_functions = ... # type: List[LambdaFunction]
    handler_map = ... # type: Dict[str, Callable[..., Any]]
    # Used for feature flag validation
    _features_used = ... # type: Set[str]
    experimental_feature_flags = ... # type: Set[str]
    FORMAT_STRING = ... # type: str

    def __init__(self, app_name: str, debug: bool=...,
                 configure_logs: bool=...,
                 env: Optional[Dict[str, str]]=...) -> None: ...

    def __call__(self, event: Any, context: Any) -> Any: ...
    def register_blueprint(self, blueprint: Blueprint,
                           name_prefix: Optional[str] = ...,
                           url_prefix: Optional[str] = ...) -> None: ...


class ChaliceAuthorizer(object):
    name = ... # type: str
    func = ... # type: Callable[[AuthRequest], Union[AuthResponse, Dict[str, Any]]]
    scopes = ... # type: List[str]
    config = ... # type: BuiltinAuthConfig
    def with_scopes(self, scopes: List[str]) -> ChaliceAuthorizer: ...


class BuiltinAuthConfig(object):
    name = ... # type: str
    handler_string = ... # type: str
    ttl_seconds = ... # type: Optional[int]
    execution_role = ... # type: Optional[str]
    header = ... # type: str
    def __init__(self, name: str, handler_string: str,
                 ttl_seconds: Optional[int] = ...,
                 execution_role: Optional[str] = ...,
                 header: str = ...) -> None: ...


class AuthRequest(object):
    auth_type = ... # type: str
    token = ... # type: str
    method_arn = ... # type: str

    def __init__(self, auth_type: str, token: str, method_arn: str) -> None: ...


class AuthRoute(object):
    path = ... # type: str
    methods = ... # type: List[str]

    def __init__(self, path: str, methods: List[str]) -> None: ...


class AuthResponse(object):
    ALL_HTTP_METHODS = ... # type: List[str]
    routes = ... # type: List[Union[str, AuthRoute]]
    principal_id = ... # type: str
    context = ... # type: Optional[Dict[str, str]]

    def __init__(
        self,
        routes: List[Union[str, AuthRoute]],
        principal_id: str,
        context: Optional[Dict[str, str]] = ...
    ) -> None: ...

    def to_dict(self, request: AuthRequest) -> Dict[str, Any]: ...


class ScheduleExpression(object):
    def to_string(self) -> str: ...


class Rate(ScheduleExpression):
    MINUTES = ...  # type: str
    HOURS = ...  # type: str
    DAYS = ...  # type: str
    value = ... # type: int
    unit = ... # type: str

    def __init__(self, value: int, unit: str) -> None: ...
    def to_string(self) -> str: ...


class Cron(ScheduleExpression):
    minutes = ... # type: Union[str, int]
    hours = ... # type: Union[str, int]
    day_of_month = ... # type: Union[str, int]
    month = ... # type: Union[str, int]
    day_of_week = ... # type: Union[str, int]
    year = ... # type: Union[str, int]

    def __init__(
        self,
        minutes: Union[str, int], hours: Union[str, int],
        day_of_month: Union[str, int], month: Union[str, int],
        day_of_week: Union[str, int], year: Union[str, int]
    ) -> None: ...
    def to_string(self) -> str: ...


class LambdaFunction(object):
    name = ... # type: str
    handler_string = ... # type: str
    func = ... # type: Callable[..., Any]


class BaseEventSourceConfig(object):
    name = ... # type: str
    handler_string = ... # type: str
    def __init__(self, name: str, handler_string: str) -> None: ...


class S3EventConfig(BaseEventSourceConfig):
    bucket = ... # type: str
    events = ... # type: List[str]
    prefix = ... # type: str
    suffix = ... # type: str


class SNSEventConfig(BaseEventSourceConfig):
    topic = ... # type: str


class SQSEventConfig(BaseEventSourceConfig):
    queue = ...                              # type: Optional[str]
    queue_arn = ...                          # type: Optional[str]
    batch_size = ...                         # type: int
    maximum_batching_window_in_seconds = ... # type: int

    def __init__(
        self, name: str, handler_string: str, queue: Optional[str],
        queue_arn: Optional[str], batch_size: int, maximum_batching_window_in_seconds: int,
    ) -> None: ...


class ScheduledEventConfig(BaseEventSourceConfig):
    schedule_expression = ...  # type: Union[str, ScheduleExpression]
    description = ...  # type: str


class CloudWatchEventConfig(BaseEventSourceConfig):
    event_pattern = ...  # type: Dict[str, Any]


class KinesisEventConfig(BaseEventSourceConfig):
    stream = ...                             # type: str
    batch_size = ...                         # type: int
    starting_position = ...                  # type: str
    maximum_batching_window_in_seconds = ... # type: int


class DynamoDBEventConfig(BaseEventSourceConfig):
    stream_arn = ...                         # type: str
    batch_size = ...                         # type: int
    starting_position = ...                  # type: str
    maximum_batching_window_in_seconds = ... # type: int


class Blueprint(DecoratorAPI):
    current_request = ... # type: Request
    lambda_context = ... # type: LambdaContext
    current_app = ... # type: Chalice
    log = ... # type: logging.Logger

    def __init__(self, import_name: str) -> None: ...

    def register(self, app: Chalice, options: Dict[str, Any]) -> None: ...

    def register_middleware(self, func: Callable,
                            event_type: str = ...) -> None: ...


class ConvertToMiddleware:
    def __init__(self,
                 lambda_wrapper: Callable[..., Any]) -> None: ...


class MiddlewareHandler(object):
    handler = ... # type: Callable[..., Any]
    next_handler = ... # type: Callable[..., Any]

    def __init__(
        self, handler: Callable[..., Any], next_handler: Callable[..., Any]
    ) -> None: ...
    def __call__(self, request: Any) -> Any: ...


class BaseLambdaHandler(object):
    def __call__(self, event: Any, context: Any) -> Any: ...
    def _build_middleware_handlers(
        self, handlers: List[Callable[..., Any]], original_handler: Callable[..., Any]
    ) -> MiddlewareHandler: ...


class RestAPIEventHandler(BaseLambdaHandler):
    api = ... # type: APIGateway
    routes = ... # type: Dict[str, Dict[str, RouteEntry]]
    debug = ... # type: bool
    log = ... # type: logging.Logger
    current_request = ... # type: Optional[Request]
    lambda_context = ... # type: Optional[LambdaContext]
    _middleware_handlers = ... # type: Optional[List[MiddlewareHandler]]

    def __init__(
        self,
        route_table: Dict[str, Dict[str, RouteEntry]],
        api: APIGateway,
        log: logging.Logger,
        debug: bool,
        middleware_handlers: Optional[List[MiddlewareHandler]]
    ) -> None: ...
    def _global_error_handler(
        self, event: Any, get_response: Callable[..., Any]
    ) -> Response: ...
    def create_request_object(self, event: Any, context: Any) -> Optional[Request]: ...
    def __call__(self, event: Any, context: Any) -> Any: ...
    def _main_rest_api_handler(self, event: Any, context: Any) -> Response: ...
    def _validate_binary_response(
        self, request_headers: Dict[str, str], response_headers: CaseInsensitiveMapping
    ) -> bool: ...
    def _get_view_function_response(
        self, view_function: Callable[..., Any], function_args: Dict[str, Any]
    ) -> Response: ...
    def _unhandled_exception_to_response(self) -> Response: ...
    def _validate_response(self, response: Response) -> None: ...
    def _cors_enabled_for_route(self, route_entry: RouteEntry) -> bool: ...
    def _get_cors_headers(self, cors: CORSConfig) -> Dict[str, Any]: ...
    def _add_cors_headers(self, response: Response, cors_headers: Dict[str, str]) -> None: ...


class EventSourceHandler(BaseLambdaHandler):
    func = ... # type: Callable[..., Any]
    event_class = ... # type: Any
    handler = ... # type: Optional[Callable[..., Any]]
    middleware_handlers = ... # type: List[Callable[..., Any]]
    _middleware_handlers = ... # type: List[Callable[..., Any]]

    def __init__(
        self,
        func: Callable[..., Any],
        event_class: Any,
        middleware_handlers: Optional[List[Callable[..., Any]]]
    ) -> None: ...
    def __call__(self, event: Any, context: Any) -> Any: ...


class BaseLambdaEvent(object):
    _event_dict = ... # type: Dict[Any, Any]
    context = ... # type: Optional[Dict[str, Any]]

    def __init__(self, event_dict: Dict[str, Any], context: Any) -> None: ...
    def _extract_attributes(self, event_dict: Dict[str, Any]) -> None: ...
    def to_dict(self) -> Dict[str, Any]: ...


class LambdaFunctionEvent(BaseLambdaEvent):
    event = ... # type: Dict[str, Any]
    context = ...  # type: Optional[Dict[str, Any]]

    def __init__(self, event_dict: Dict[str, Any], context: Any) -> None: ...


class CloudWatchEvent(BaseLambdaEvent):
    version = ... # type: str
    account = ... # type: str
    region = ... # type: str
    detail = ... # type: Dict[str, Any]
    detail_type = ... # type: str
    source = ... # type: str
    time = ... # type: str
    event_id = ... # type: str
    resources = ... # type: List[str]


class SQSRecord(BaseLambdaEvent):
    body = ... # type: str
    receipt_handle = ... # type: str


class SQSEvent(BaseLambdaEvent):
    def __iter__(self) -> Iterator[SQSRecord]: ...


class SNSEvent(BaseLambdaEvent):
    message = ... # type: str
    subject = ... # type: str


class S3Event(BaseLambdaEvent):
    bucket = ... # type: str
    key = ... # type: str


class KinesisRecord(BaseLambdaEvent):
    data = ... # type: bytes
    sequence_number = ... # type: str
    partition_key = ... # type: str
    schema_version = ... # type: str
    timestamp = ... # type: datetime.datetime


class KinesisEvent(BaseLambdaEvent):
    def __iter__(self) -> Iterator[KinesisRecord]: ...


class DynamoDBRecord(BaseLambdaEvent):
    @property
    def table_name(self) -> str: ...

    timestamp = ... # type: datetime.datetime
    keys = ... # type: Any
    new_image = ... # type: Any
    old_image = ... # type: Any
    sequence_number = ... # type: str
    size_bytes = ... # type: int
    stream_view_type = ... # type: str
    aws_region = ... # type: str
    event_id = ... # type: str
    event_name = ... # type: str
    event_source_arn = ... # type: str


class DynamoDBEvent(BaseLambdaEvent):
    def __iter__(self) -> Iterator[DynamoDBRecord]: ...


class MultiDict(MutableMapping):
    _dict = ...  # type: Dict[Any, Any]

    def __init__(self, mapping: Dict[Any, Any]) -> None: ...
    def __getitem__(self, k: str) -> Optional[Any]: ...
    def __setitem__(self, k: str, v: Any) -> None: ...
    def __delitem__(self, k: str) -> None: ...
    def getlist(self, k: str) -> List[Any]: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Any: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...


class CaseInsensitiveMapping(Mapping):
    _dict = ... # type: Dict[Any, Any]

    def __init__(self, mapping: Dict[Any, Any]) -> None: ...
    def __getitem__(self, key: str) -> Any: ...
    def __iter__(self) -> Any: ...
    def __len__(self) -> int: ...
    def __repr__(self) -> str: ...


class WebsocketEvent(BaseLambdaEvent):
    domain_name = ... # type: str
    stage = ... # type: str
    connection_id = ... # type: str
    body = ... # type: str


unquote_str = ... # type: Callable[..., Any]
